\ Strand - interprocess communication

.( ipc )

256 constant MAX_PORTS      

begin-structure /port
  field: port.lock   field: port.owner      field: port.size   
  PORT_SIZE 3 cells - +field port.data
end-structure
2variable msgfile       variable msgbase    
variable msgport        variable eventloop
variable sent       variable received
variable #sent       variable #received
variable %events       variable %events_head
variable %send_read     variable %send_get_module   
variable %send_value   variable %send_lookup
variable %peers       

: gcipc  %events gctrace!  %events_head gctrace! 
  %peers gctrace! ;
: addr>machine  ( addr -- mid ) 8 rshift ;
: addr>pix  ( addr -- pix ) 255 and ;
: .addr  ( addr -- ) dup addr>machine (.) type  [char] : emit  
  addr>pix . ;
: portdata  ( port -- a u ) dup port.data swap port.size @ ;
: local-ns  ( -- addr ) machine-id @ 8 lshift ;
: tunneled?  ( addr -- f ) addr>machine  machine-id @ <> ;
: localaddr  ( addr -- pix ) 
  dup addr>machine  machine-id @  =  if
    addr>pix  
  else
    addr>machine  
  then ;
: >port  ( pix -- port ) PORT_SIZE * msgbase @ + ;
[defined] arm  [if]
: trylock  ( port -- f ) port.lock 1 lock? ;
[else]
: trylock  ( port -- f ) port.lock 1 0 rot cas ;
[then]
: lock  ( port -- ) begin  dup trylock 0=  while
    pause  repeat  drop  busy  ;
: unlock  ( port -- ) port.lock off ;
: emptyport  ( port -- ) port.size off ;
: unown  ( port -- ) port.owner off ;
: setport  ( pix -- ) dup portx !  >port dup msgport !  
  port.owner node-id @ swap !  
  <log  ." message port index is "  portx ?  cr  log> ;
: portfull?  ( port -- f ) port.size @ ;
: portempty?  ( port -- f ) port.size @ 0= ;
: lockempty  ( port -- f ) \ return false on fail
  LOCK_ATTEMPTS swap  begin  over  while
    dup portempty?  if
      dup lock  dup portempty?  if  2drop  true  |
      dup unlock
    then 
    pause  -1 under+ 
  repeat 
  [if-debug-build]
    <dlog  ." too many lock-attempts - yielding ..."  cr  log>
  [then]
  2drop  busy  false ;
: claimport  node-id @ MAX_PORTS 1- and 1 or  
  begin  
    255 and dup >port dup port.owner @ 0=  if   ( pix port )
      dup lock  dup port.owner @ 0=  if  swap setport  unlock  |
      unlock
    then  drop
    1+  255 >  if  1  then again ;
: exitmsg  msgport @ ?dup  if  dup emptyport  unown  then ;
: nomsgfile 
  <log  ." can not open message-file: "  msgfile 2@ type  cr  
  log> ;
: initport  portx @ dup -1 <>  if  setport  
  else  drop  claimport  then ;
: openmsgfile  
  msgfile 2@ r/w map-file if  nomsgfile  |
  drop msgbase !  drop
  <log  ." node " node-id ?  ." opened message file " 
    msgfile 2@ type  ."  at "  msgbase ?  cr  log>
  initport ;
: initmsg  initrbuckets
  msgbase @ 0=  if  openmsgfile  |  
  <log  ." node " node-id ?  ." uses message file address " 
    msgbase ?  cr  log>
  initport ;
: initevents 
  make_var dup %events !  %events_head ! ;
: ?msg  msgbase @ 0=  if  
    <fatal  ." no message file open"  fatal>  then ;

: (add_event)  ( x -- ) 
  make_var dup >r make_list  %events @ swap assign  
  r> %events ! ;
' (add_event) is add_event

variable recvr      variable sleeping
: ns-wait  ( localaddr -- ) ?exit  0 >port
  sleeping off
  begin  dup port.owner @ 0=  while
      pause  interrupts  sleeping @ 0=  if  
        sleeping on  <log  ." waiting for name server ..."  cr  log>  then
  repeat  drop  busy ;
: ?owner  ( port -- ) port.owner @ 0=  if
    <fatal  ." port has no owner: "  recvr ?  fatal>  then ;

: send  ( x addr -- ) ?msg  ?dup 0=  if  local-ns  then
  dup recvr !  localaddr  dup ns-wait   ( x pix )
  >port dup ?owner  dup lockempty 0=  if  yield  |  ( x port )
  dup /port m-start  dup port.data dup >r   ( x port pdata )
  recvr @ tunneled?  if  recvr @ m-pack  then
  rot  marshal  m-end  r> - dup sent +!  over port.size !   ( port )
  <log  .clock ." sent to "  recvr @ .addr ." <<"  
  dup portdata tuck type  ." >> "  .  cr  log>  unlock 
  1 #sent +! ;

\ special simplified case, used for $deliver/1
: send_data  ( a u locaddr -- ) dup recvr !  
  addr>pix >port  dup ?owner 
  dup >r lockempty 0=  if  r> drop  yield  |
  over sent +!
  tuck r@ port.data swap cmove  r@ port.size !  r>
  <log  .clock ." sent forwarded data to "  recvr @ .addr ." <<"  
  dup portdata type  ." >>"  cr  log>  unlock ;

: (receive)  ( -- f ) 
  msgport @ dup portfull? 0=  if  drop  false  |  
  dup portdata
  <log  .clock  ." received on port "  portx ?  ." <<"  
    2dup type  ." >> "  dup .  cr  log> 
  dup received +!  dup cells ?heap
  um-start  dup port.data unmarshal  um-end
  nip  add_event  dup emptyport  unlock  1 #received +! 
  true ;
: receive  msgport @ 0= ?exit  eventloop @ 0= ?exit
  sleeping off 
  begin (receive) ?exit  qempty?  while  
      pause  interrupts  0 poll_input 
      sleeping @ 0=  if  
        sleeping on  
        [if-debug-build] 
          <dlog  ." listening"  cr  log>  
        [then]
      then
  repeat  busy ;
' receive is msg_hook

: getremote  ( rem -- var ) 
  [if-debug-build]
    <dlog  ." getremote: " dup write  cr  log>
  [then]
  dup rem.var @ ?dup  if  nip  |
  push  12 cells ?heap
  top rem.owner @  top rem.id @
  %send_read @ 3 make_tuple add_event
  make_var dup pop rem.var ! ;
' getremote is remoteref 
: setremote  ( x rem -- ) 
  [if-debug-build]
    <dlog  ." assigning "  over dwrite  ."  to remote "  
      dup write  cr  log>
  [then]
  dup rem.id @ >r  dup rem.owner @ >r
  swap push  rem>var  dup  top assign
  /var 12 cells + ?heap  \ + var/list for event
  pop r> r> %send_value @  4 make_tuple add_event ;
' setremote is remoteassign

: request_module  ( nid mid var -- ) 
  push  push  >r  5 cells ?heap  pop pop swap r> 
  %send_get_module @ 4 make_tuple add_event ;

: request_peer  ( name -- var ) /var 8 cells + ?heap
  make_var  dup >r  swap  %send_lookup @  3 make_tuple
  add_event  r> ;
: resolve_peer  ( name -- pidx|var ) >r
  %peers @  begin  %[] @  ->  r> request_peer  |
    dup car dup car r@ =  if  r> drop  cdr  nip  |
    drop  cdr  
  again ;
: add_peer  ( peer pidx -- ) 
  6 cells ?heap  make_list  %peers @ make_list %peers ! ;
