\ Strand - library (builtin processes)

.( lib )

4 constant EINTR

: tohere  ( a -- a u ) dup here swap -  over h ! ;
: listlength  ( lst -- u ) 0 swap  
  begin  deref dup list?  while
    1 under+ cdr  repeat  
  drop ;
: index  ( tup u -- a ) 2dup  swap size 0 swap within 0=
  if  <err" range"  ." tuple index out of range: "  1+ .  err>  |  th ;
: atomlength  ( atm -- u ) 0 swap bdata over +  
  begin 2dup <  while  swap utfdecode drop swap 
    rot 1+ -rot  repeat  
  2drop ;
variable %send_call

\ builtins
: $reclaim/2  ( vf vt -- ) !var push  !var push
  gcreclaim  pop gcfree integer assign  pop gctotal integer assign ;
  0 builtin

: $events/1  ( v -- ) 
  !var %events_head @ assign  eventloop on
  %events_head off  ; 1 builtin
: $send/2  ( pix val -- ) ?deref  swap ?int send ; 2 builtin
: $send/3  ( pix val v -- ) !var -rot  $send/2  %[] @ assign ;
  3 builtin
: $node/1  ( var -- ) !var self integer assign ; 4 builtin

: assignrem  ( x rem -- ) dropremote
  dup rem.var @ dup 0= abort" assigned remote has no variable"
  tuck mutate  swap assign ;  
: $assign_remote/3  ( id owner x -- ) 
  [if-debug-build]
    <dlog  ." assign_remote: " 3 writestack  log>
  [then]
  -rot swap get_int swap get_int  (findremote) \ assigns `id/owner`
  ?dup 0=  if  drop  <dlog  ." remote not found - ignored"  cr  log>  |
  dup remote?  if  assignrem  |  
  swap assign  id/owner 2@ dropref ; 5 builtin
: $get_exposed/2  ( id var -- ) 
  [if-debug-build]
    <dlog  ." get_exposed: " over write space  dup write cr  log>
  [then]
  swap get_int localremote assign ; 6 builtin

: $log/1  ( x -- ) deref*  <log  dup write  cr  log>  drop ; 
  7 builtin

: rcall  ( mid pdix -- ) >r  push 6 cells ?heap
  5 (make_tuple) %send_call @ over !  pop over 1 th !
  r> integer over 2 th !  pop over 3 th !  pop over 4 th !  add_event 
  terminate ;
: rcall_external  ( mname name arity -- ) 
  >r push resolve_module
  dup pop r> rot find_pdix  swap mod.id @ swap rcall ;
: $call_remote/4  ( mid|mname pidx|name tup node -- )
  <dlog  ." call_remote: " 4 writestack  log>
  push push  dup integer?  if  get_int  rcall  |
  top size rcall_external ; 8 builtin

: $call/3  ( mod pidx tup -- ) !tup push  ?int swap !mod
  mod.pdefs @ swap th @ r> drop >r  pop 0 push_args ; 
  9 builtin
: $retrieve_module/3  ( mid nid var -- ) !var push  !int
  swap !str dup find_mid ?dup 0=  if  pop request_module  |
  pop swap assign  2drop ; 10 builtin
: $load_module/2  ( mid var -- ) !var push  !str  dup
  find_mid ?dup 0=  if  <err" load"  ." module not found: "  write  err>  |
  nip  pop swap assign ; 11 builtin
: ?module  ( x -- mod )
  dup atom?  if  resolve_module  then  !mod ;
variable %/
: module_exports/2  ( mod var -- ) !var push  ?module
  mod.exports @ dup push size dup 2* cells ?heap \ roughly
  %[] @  swap  0  ?do
    top i 1+ th @  top i th @  %/ @  3 make_tuple  swap make_list
  3 +loop
  pop drop  pop swap assign ; 77 builtin

: halt/1  ( i -- ) ?int  stopnode  
  <log  ." node terminated with exit code " dup .  cr  log>
  exitcode ; 12 builtin
: error/1  ( x -- ) deref* dup %failure !  
  \ fake <err" ... err>
  (<err)  [ here  0 , ]  write  [ here swap ! ]  (err>) ; 13 builtin

: write/1  ( x -- ) deref* write ; 14 builtin
: writeln/1  ( x -- ) write/1  cr ; 15 builtin

: bytelist  ( a u tail -- lst ) push  dup 4 * cells ?heap  \ N pairs
  tuck +  swap  0  ?do  
    1- dup c@ integer pop make_list push  loop
  drop  pop ;
: charlist  ( a u tail -- x ) push dup 4 * cells ?heap  \ N pairs
  here >r  over +  
  begin  2dup <  while
    swap utfdecode ,  swap  repeat  
  2drop  
  here dup r> - bytes  0  ?do  
    1 cells - dup @ integer pop make_list push  loop
  h !  pop ; 
32 k constant SAREA_OFF
variable >sarea
: sarea  ( -- a ) here SAREA_OFF + ;
: sreset  sarea >sarea ! ;
: sc,  ( c -- ) >sarea @ c!  1 >sarea +! ;
: sutfencode  ( u -- ) >sarea @ swap utfencode >sarea +!
  drop ;
: stringify  ( x -- a u ) >sarea @ >r
  begin  %[] @  ->  0 sc,  r> dup >sarea @ diff  |
    !list @+ ?int sutfencode  @ deref
  again ;
: ?str  ( x -- a u ) deref dup atom?  if  bdata  |
  dup list? 0=  if  s" string or list" expected  |
  stringify ;

: string_to_list/3  ( s v x -- ) swap !var push  push
  !str bdata pop charlist  pop swap assign ; 16 builtin
: list_to_string/2  ( lst v -- ) !var push  here >r
  begin  deref %[] @  ->  pop r> tohere intern assign  |
    !list @+ ?int here swap utfencode allot drop  @  
  again ; 17 builtin

: make_tuple/2  ( i v -- ) !var push  ?int dup
  (make_tuple)  push  
  0  ?do  make_var top i th !  loop  
  pop  pop swap assign ; 18 builtin

variable ibase      10 ibase !
: stdbase  10 ibase !  decimal ;
: non_int  ( a u -- )
  <err" type"  ." can not convert to integer: "  type  err> ;
: list_to_integer/2  ( l v -- ) !var swap  !list  here >r  
  begin  
    deref %[] @  ->  r> tohere  
      ibase @ base !  number  stdbase  0=  if  non_int  |
      integer assign  |  
    dup car ?int c, cdr  
  again ; 19 builtin
: integer_to_list/2  ( i v -- ) !var push  %[] @ push  ?int
  ibase @ base !  (.)  stdbase
  dup 4 cells * ?heap  tuck + 
  swap  0  do  
    1- dup c@ integer pop make_list push  loop  
  drop  pop pop swap assign ; 20 builtin
variable ltmp
: list_to_tuple/2  ( lst var -- ) !var push
  deref dup null?  if  
    drop  0 (make_tuple) 
  else
    !list
    dup listlength swap push  dup 4 cells * ?heap
    dup (make_tuple) ltmp !  pop 
    swap  0  do  dup car ltmp @ i th !  cdr deref  loop
    ltmp @ then
  pop swap assign ; 21 builtin
: tuple_to_list/3  ( tup var x -- ) 
  swap !var push  swap !tup dup size >r
  push  push  r@ 4 cells * ?heap
  pop  pop  r@ th  
  r>  0  ?do  1 cells - dup @ rot make_list  swap  loop  
  drop  pop swap assign ; 22 builtin

: length/2  ( x v -- ) !var swap deref
  dup null?  if  drop  0integer assign  |
  dup atom?  if  atomlength integer assign |
  dup list?  if  listlength integer assign |  
  size integer assign ; 23 builtin

: run/2  ( mod x -- ) deref dup >r functor rot deref
  ?module find_export  r> swap >r 1 push_args ; 24 builtin
: get_module/2  ( atm var -- ) !var push  !str
  resolve_module pop swap assign ; 25 builtin
: newtask  task_drop  make_task %me @ process.task ! ;
: run/4  ( mod term svar ctl -- ) 
  2swap  deref push
  deref push  push  !var push  
  256 ?heap  \ task (+ 2 vars)
  newtask  pop task task.status !  pop task task.control !
  <dlog  .task  ." created"  cr  log>
  pop ?module  pop dup >r functor rot find_export
  r> swap >r  1 push_args ; 69 builtin
: $run/5  ( mid pidx tup svar ctl -- )
  push  !var push  deref push  ?int >r  
  deref find_mid ?dup 0= abort" module not available"
  push  256 ?heap  \ task (+ 2 vars)
  pop  newtask  pop  ( mod tup )
  pop task task.status !  pop task task.control !
  <dlog  .task  ." created"  cr  log>
  swap mod.pdefs @ r> th @ r> drop >r
  0 push_args ; 73 builtin

: assign/3  ( var1 x var2 -- ) !var push  
  swap !var swap assign  pop %[] @ assign ; 26 builtin

: put_arg/3  ( int tup x -- ) deref push  !tup  swap ?int 1-
  index @ pop assign ; 27 builtin
: put_arg/4  ( int tup x var -- ) !var push  put_arg/3  pop 
  %[] @ assign ; 28 builtin
: get_arg/3  ( int tup var -- ) !var push  !tup  swap ?int 1-
  index @ pop swap assign ; 29 builtin
: get_arg/4  ( int tup var1 var2 -- ) !var push  get_arg/3  pop
  %[] @ assign ; 30 builtin

: iomode  ( atm -- m ) bdata
  2dup s" rw" compare 0=  if  2drop  r/w  |
  2dup s" r" compare 0=  if  2drop r/o  |
  2dup s" w" compare 0=  if  2drop w/o  |
  2dup s" a" compare 0=  if  2drop a/o  |
  <err" open"  ." invalid I/O mode: "  type err> ;
: open_file/3  ( atm dir var -- ) !var push  !str iomode swap 
  sreset dup >r ?str rot open-file  if  
    <err" open"  ." error while opening file: "  r> write  err>  |
  integer pop swap assign  r> drop ; 31 builtin
: close_file/1  ( fd -- ) ?int dup close-file  if  
    <err" close"  ." error while closing file: "  .  err>  |  drop ; 32 builtin
: readfailed  ( fd -- )
  errno @ EINTR =  if  drop  yield  |
  <err" read"  ." error while reading from file: "  .  err> ;
: read_bytes/4  ( fd len var x -- ) swap !var push  push
  ?int  >r  ?int dup here r> rot read-file  if  drop  readfailed  |
  nip  ?dup 0=  if  pop drop  pop %[] @ assign  |
  here dup >r swap dup allot pop bytelist  r> h !  pop swap assign ; 
  33 builtin
variable wfd
: dowrite  ( a u -- ior ) 
  begin  2dup wfd @ write-file  0  ->  2drop  0  |
    errno @ EINTR <>  if  nip nip  |
  again ;
: fwritelist  ( lst -- ior ) here >r
  begin  %[] @  ->  r@ here r@ - dowrite  r> h !  |
    !list dup car ?int here swap utfencode allot drop  
    cdr deref  again ; 
: writefailed  ( fd -- )
  errno @ EINTR =  if  drop  yield  |
  <err" write"  ." error while writing to file: "  .  err> ;
: write_chars/2  ( fd lst -- )  swap ?int wfd ! 
  deref fwritelist  if  wfd @ writefailed  then ; 34 builtin
: write/2  ( fd x -- )  deref* swap ?int  stdout @ >r stdout !
  write  r> stdout ! ; 35 builtin

\ merger: create merger process for each added merge(_) msg
\ destructively modified cell holds count of attached merger
\ processes
variable %merge
: assign_tail  ( lst x -- ) deref  swap  begin  
    ?deref dup var? 0=  while  !list cdr  repeat  swap assign ;
: merger_close  ( str rcell -- )
  deref dup @ get_int  1  ->  drop  %[] @ assign_tail  |
  1- integer swap !  drop ;
: merge?  ( x -- f ) dup tuple? 0=  if  drop  false  |
  dup size 2 <>  if  drop  false  |
  @ %merge @ = ;
defer merger1
: mergeproc  %me @ process.args @ @+ swap @+ swap @
  merger1  schedule ;
: mkmerger  ( rcell str2 str1 -- ) 3 make_tuple
  ['] mergeproc >body make_process tuck process.args !  
  %merge @ over process.loc !  enqueue ;
: mergepush  ( str2 rcell str1 x -- ) 
  make_var dup >r make_list ( str2 rcell str1 str2' ) rot >r swap >r
  assign_tail  r> cdr  r>  r> ( str1' rcell str2' ) rot mkmerger ;
: addmerge  ( str2 rcell str1 str3 -- )
  push  push  deref 2 over +! push  push  
  pop dup >r  pop dup >r  2nd  ( str2 rcell' str3 ) rot swap mkmerger
  pop cdr  r> r> ( str1' rcell' str2 ) rot  pop  drop mkmerger ;
: (merger1)  ( str1 str2 rcell -- ) 
  rot deref  %[] @  ->  merger_close  |
  !list dup car dup merge?  if  1 th @ addmerge  |  mergepush ;
' (merger1) is merger1
: merger/2  ( str1 str2 -- ) push push  1 integer 1 make_tuple 
  pop pop swap mkmerger ; 36 builtin

: blist_to_integer/3  ( l i v -- ) rot deref* -rot 
  swap ?int ibase !  list_to_integer/2 ; 37 builtin
: binteger_to_list/3  ( i1 i2 v -- ) rot deref -rot 
  swap ?int ibase !  integer_to_list/2 ; 38 builtin

: byte_length/2  ( x v -- i ) !var swap !block
  octets integer assign ; 39 builtin

: write/3  ( fd x v -- ) !var push  write/2  pop %[] @ assign ;
  40 builtin
: write_chars/3  ( fd lst var  -- ) !var push  write_chars/2  
  pop %[] @ assign ; 41 builtin

: string_to_byte_list/3  ( s v x -- ) swap !var push  push
  !str bdata pop bytelist  pop swap assign ; 42 builtin

: fwriteblist  ( lst -- ior ) here >r
  begin  %[] @  ->  r@ here r@ - wfd @ write-file  r> h !  |
    !list dup car ?int c,  cdr deref  again ; 
: write_bytes/2  ( fd lst -- )  swap ?int wfd ! 
  deref fwriteblist  if  wfd @ writefailed  then ; 43 builtin
: write_bytes/3  ( fd lst var  -- ) !var push  write_bytes/2  
  pop %[] @ assign ; 44 builtin

: command_line/1  ( var -- ) !var push  #arg 1-  %[] @  begin
    over 0 >= while  over arg intern swap make_list  -1 under+
  repeat  nip  pop swap assign ; 45 builtin

: $resolve_peer/2  ( peer var -- ) !var push
  dup integer? 0=  if  !str resolve_peer  then  
  pop swap assign ; 46 builtin

: decode1  ( lst int -- lst' int' ) swap deref  %[] @  ->  %[] @ swap  |
  !list dup cdr swap car deref ?int 63 and  rot  6 lshift or ;
: decode  ( {v2 v1} lst int -- )
  dup 32 and  if
    dup 16 and  if
      7 and decode1 decode1 decode1
    else
      15 and decode1 decode1
    then
  else
    31 and decode1                       
  then
  integer pop swap assign  pop swap assign ;
: utf_decode/3  ( lst v1 v2 -- ) !var push  !var push
  !list dup cdr swap car ?int dup 128 and  if  decode  |
  integer  pop swap assign  pop swap assign ; 47 builtin

: $register_peer/2  ( peer pidx -- ) !int swap  !str  swap
  add_peer ; 48 builtin
: $node/2  ( var1 var2 -- ) !var node-id @ integer assign
  !var self integer assign ; 49 builtin
: $add_reference/1  ( id -- ) ?int  self addref ; 
  50 builtin
: $drop_reference/1  ( id -- ) ?int  self dropref ;
  51 builtin

: listens  ( fdint -- var|0 )
  %listening @  
  begin  dup %[] @ <>  while
    2dup car car =  if  car cdr  nip nip  |
    cdr
  repeat  drop  0 ;
: listen/2  ( fd var -- ) 
  !var push !int  6 cells ?heap
  dup listens ?dup  if  nip  pop swap assign  |
  pop make_list
[if-debug-build]
  <dlog  ." listen: " dup dwrite  cr  log>
[then]
  %listening @ make_list %listening ! 
  relisten on ; 52 builtin

: $forward/2  ( fd pack -- ) stdout @ >r  swap ?int stdout !
  !tup  here PORT_SIZE m-start
  <log  ." forwarding: " dup write  cr  log>
  here swap marshal  m-end  here - dup 1+ (.) tuck type  8 diff spaces
  here swap type  cr
  r> stdout ! ; 53 builtin
: readerr  <fatal  ." error while accepting forwarded message"  
  fatal> ;
: readfill  ( a u fd -- f ) >r 
  begin  ?dup  while
    2dup r@ read-file ?ior ?dup 0=  if  2drop  r> drop  false  |
    /string  repeat  r> 2drop  true ;
: readmsg  ( fd -- tup ) 
  >r  here 8 r@ readfill 0=  if  
    <log  ." EOF from remote peer"  cr  log>  stopnode  0 exitcode  |
  here 8 bl scan  nip 8 diff  here swap  number 0=  if  readerr  |
  ?dup 0=  if  <log  ." remote peer indicated end of communication"  cr
    log>  stopnode  0 exitcode  |
  PORT_SIZE ?heap
  here swap 2dup um-start  r> readfill 0=  if
    <fatal  ." unexpected end of input from remote peer"  fatal>  |
  here unmarshal nip  um-end ;
: $read_forwarded  ( fd var -- ) swap ?int readmsg assign ;
  64 builtin
: $deliver/1  ( msg -- ) !tup
  <log  ." delivering: " dup write  cr  log>
  dup 1 th @ bdata  rot @ get_int  send_data ; 54 builtin

: strand_version/1  ( var -- ) !var strand-version integer assign ;
  55 builtin

: randomize/1  ( int -- ) ?int seed ! ; 56 builtin

: $statistics/1  ( var -- ) !var push  14 cells ?heap
  #proc integer  #spcount @ integer  #derefs @ integer
  #s-total @ integer  #r-total @ integer  sent @ integer
  #sent @ integer  received @ integer  #received @ integer
  #remotes @ integer  #exposed @ integer #atoms @ integer
  staticspace integer  13 make_tuple  pop swap assign ;
  57 builtin
: time/1  ( var -- ) !var time integer assign ; 75 builtin

: close_file/2  ( fd var -- ) !var swap  close_file/1
  %[] @ assign ; 58 builtin

variable %''        variable %os        variable %arch
: platform/2  ( var var -- ) !var %arch @ assign  
  !var %os @ assign ; 59 builtin
: getenv/2  ( str var -- ) !var  swap sreset ?str getenv
  ?dup  if  intern assign  |  %'' @ assign ; 60 builtin

: shell/2  ( str var -- ) !var  swap
  <log  ." spawning child process: "  dup dwrite  cr  log>
  sreset ?str
[defined] handle  [if]
  <fork  shell  fork>  nip nip addchild 
[else]
  system integer assign
[then] 
  ; 61 builtin
: open_pipe/2  ( var1 var2 -- ) !var swap  !var
  pipe >r integer assign  r> integer assign ; 
  62 builtin
: redirect  ( fd1 fd2 -- ) swap  -1  ->  drop  |
  2dup =  if  2drop  |  swap (>[]) ;
: closelist  ( fdlst -- ) begin  dup %[] @ <>  while
    dup car ?int close  cdr  
  repeat  drop ;
variable xargs
: execargs  ( lst -- a ) 
  dup @ ?str over c@ [char] / <>  if
    searchpath drop
  then
  here xargs !  sreset
  zstring ,  cdr
  begin  dup %[] @ <>  while
    dup car ?str drop ,  cdr 
  repeat  
  drop  0 ,  xargs @ ;
: !strlist  ( x -- x ) deref*  dup
  begin  dup %[] @ <>  while  !list dup  car !str/list drop  cdr  repeat
  drop ;
: $execute/7  ( lst svar fin fout ferr clst var -- )
  !var push  deref* push  ?int >r  ?int >r  ?int >r  !var >r 
  !list  !strlist
  <log  ." spawning child process: "  dup dwrite  cr  log>
  r>  r> r> r>  ( svar fin fout ferr )
  <fork  2 redirect  1 redirect  0 redirect
    drop  pop closelist
    execargs exec  <fatal  ." execve(2) failed" fatal>
    fork>  pop drop  pop over integer assign  ( fin fout ferr pid )
  >r  2drop  drop  r> addchild  drop ; 63 builtin

: check_ior  ( ior -- ) 0= ?exit  
  <err" io"  ." error in I/O operation" err> ;
: file_size/2  ( str var -- )
  !var swap  sreset ?str r/o open-file check_ior dup file-size ?ior
  swap close  integer assign ; 65 builtin
: file_modification_time/2  ( str var -- )
  !var swap  sreset ?str r/o open-file check_ior dup file-mtime ?ior
  swap close  integer assign ; 66 builtin

: string_to_integer/2  ( str var -- )
  !var push  !str bdata  ibase @ base !  
  number  stdbase  0=  if  non_int  |
  integer  pop swap assign ; 67 builtin
: bstring_to_integer/3  ( str base var -- )
  swap ?int ibase !  string_to_integer/2 ; 68 builtin

: put_global/3  ( key val var -- ) !var push 
  swap !str put_global  pop %[] @ assign ; 70 builtin
: get_global/2  ( key var -- ) !var  swap !str get_global
  assign ; 71 builtin

: chdir/2  ( str var -- ) !var swap  sreset ?str chdir check_ior 
  %[] @ assign ; 72 builtin

[defined] linux  [if]
  [defined] arm  [if]
16 constant stat.mode
  [else]
24 constant stat.mode
  [then]
[else]
  [defined] darwin  [if]
8 constant stat.mode
  [else]
0 constant stat.mode
  [then]
[then]
61440 constant S_IFMT
16384 constant S_IFDIR      40960 constant S_IFLNK
variable %file      variable %directory     variable %link
variable %none
: file_status/2  ( str var -- ) !var  swap sreset ?str
  r/o open-file  if  
    drop  %none @  assign  |
  dup statbuf sys_fstat syscall2 drop  statbuf stat.mode + w@
  swap close
  S_IFMT and dup S_IFDIR =  if  drop  %directory @  assign  |
  S_IFLNK =  if  %link @  assign  |
  %file @  assign ; 74 builtin

: deref/2  ( x var -- ) !var  swap deref* drop  %[] @ assign ; 
  76 builtin
