% stream support library
%
% bounded_buffer(COUNT?, IN?, OUT^)
%   Creates an intermediate stream bounded to COUNT elements.

-exports([bounded_buffer/3]).

bounded_buffer(N, IN, OUT) :-
    bounded_buffer2(N, BB, BT, DONE),
    forward(IN, BB),
    bounded_buffer3(DONE, BB, BT, OUT).

bounded_buffer2(0, BH, BT, DONE) :- 
    BT := BH, DONE := [].
bounded_buffer2(N, BH, BT, DONE) :-
    N > 0 | 
    BH := [_|BT1],
    N2 is N - 1,
    bounded_buffer2(N2, BT1, BT, DONE).

bounded_buffer3([], [stop|_], _, OUT) :-
    OUT := [].
bounded_buffer3([], [X|BB], BT, OUT) :-
    data(X) |
    BT := [_|BT2],
    OUT := [X|OUT2],
    bounded_buffer3([], BB, BT2, OUT2).

forward([], [X|_]) :- X := stop.
forward([X|R], [V|BB]) :-
    data(X) |
    V := X, forward(R, BB).
