% Operations on Strand terms

-exports([args/2, vars/3, classify_arg/6]).
-exports([flatten_body/2, flatten_body/3]).

% extract term arguments from toplevel term

args((H :- _), ARGS) :- args1(H, ARGS).
args(H, ARGS) :- args1(H, ARGS).

args1(T, ARGS) :-
    tuple(T) | tuple_to_list(T, L, []), args2(L, ARGS).
args1(_, ARGS) :- otherwise | ARGS := [].

args2([_|XS], ARGS) :- ARGS := XS.

% extract variables from term

vars(T, TAG, VS) :- vars(T, TAG, [], VS).

vars('$VAR'(TAG, I, N), TAG, VS1, VS) :- 
    VS := [{I, N}|VS1].
vars([X|R], TAG, VS1, VS) :-
    vars(X, TAG, VS1, VS2), 
    vars(R, TAG, VS2, VS).
vars(T, TAG, VS1, VS) :-
    tuple(T) | 
    tuple_to_list(T, L, []), 
    vars(L, TAG, VS1, VS).
vars(_, _, VS1, VS) :- 
    otherwise | VS := VS1.

% classify process-definition clauses by 1st argument

classify_arg([], _, TL, LL, AL, OL) :-
    TL := [], LL := [], AL := [], OL := [].
classify_arg([(H :- otherwise | BODY)|R], _, TL, LL, AL, OL) :-
    TL := [], LL := [], AL := [], OL := [(H :- otherwise | BODY)|R].
classify_arg([T|R], VTAG, TL, LL, AL, OL) :-
    args(T, ARGS),
    classify_arg1(ARGS, T, R, VTAG, TL, LL, AL, OL).

classify_arg1([], T, R, _, TL, LL, AL, OL) :- 
    TL := [], LL := [], AL := [], OL := [T|R].
classify_arg1(['$VAR'(VTAG, _, _)|_], T, R, VTAG, TL, LL, AL, OL) :-
    OL := [T|OL2], 
    classify_arg(R, VTAG, TL, LL, AL, OL2).
classify_arg1([X|_], T, R, VTAG, TL, LL, AL, OL) :-
    tuple(X) | 
    TL := [T|TL2], 
    classify_arg(R, VTAG, TL2, LL, AL, OL).
classify_arg1([X|_], T, R, VTAG, TL, LL, AL, OL) :-
    list(X) | 
    LL := [T|LL2], 
    classify_arg(R, VTAG, TL, LL2, AL, OL).
classify_arg1([_|_], T, R, VTAG, TL, LL, AL, OL) :-
    otherwise | 
    AL := [T|AL2], 
    classify_arg(R, VTAG, TL, LL, AL2, OL).

% flatten conjunctions into list

flatten_body(X, B) :- flatten_body(X, B, []).

flatten_body((true, Y), B, TL) :- flatten_body(Y, B, TL).
flatten_body((X, true), B, TL) :- flatten_body(X, B, TL).
flatten_body((X, Y), B, TL) :-
    X =\= true, Y =\= true |
    flatten_body(X, B, TL1), flatten_body(Y, TL1, TL).
flatten_body(X, B, TL) :- 
    otherwise | B := [X|TL].
